module coeffs
!-------------------------------------------------------------------------------------
! Coefficient and boundary condition functions that define the general 
! 2D linear PDE:
!     au_xx + bu_yy + cu_x + du_y + eu = f,
!     subject to u = BC on the boundaries
! The boundary conditions are Dirichlet. All of a-f can be functions of x and y.
!
! Look in the subdirectory Coeffs for other coefficient sets.
! This particular set is for the Poisson equation, with boundary conditions set to
! match those of the diffusion problem covered earlier.
!
! Warning: the way the boundary conditions are set is by testing for exact equality
! between two double precision numbers. This is generally a Dumb Idea (TM), but works
! here because the numbers  (0.0 and 1.0) are exactly representable in floating point
! arithmetic. If the domain was on [-0.1, 0.1] X  [-0.1, 0.1] then some epsilon based
! comparison would need to be made instead. The gfortran compiler with -Wall option
! will squawk about such comparisons. It will also gripe about unused inputs when,
! e.g., afun = 1.0, because then x and y are not used.
!
! Another thing: the function BC(x,y) should only be called when at least one of x or
! y is equal to zero or to one. Otherwise, no value is set. I have put in a test for
! that in this coefficient set, but in general do not.
!-----------------
! Randall Bramley
! Department of Computer Science
! Indiana University, Bloomington
!-------------------------------------------------------------------------------------

    use kindlytypes, only: real8
    use iso_fortran_env, only: stderr => error_unit
    real(kind=real8), parameter :: pi = 3.1415926535897932_real8

contains 
    real(kind=real8) function afun(x,y)
        real(kind=real8) :: x, y
        afun =  1.0_real8
        return
    end function afun

    real(kind=real8) function  bfun(x,y) 
        real(kind=real8) :: x, y
        bfun = 1.0_real8
        return
    end function bfun

    real(kind=real8) function  cfun(x,y) 
        real(kind=real8) :: x, y
        cfun = 0.0_real8
        return
    end function cfun

    real(kind=real8) function  dfun(x,y) 
        real(kind=real8) :: x, y
        dfun =  0.0_real8
        return
    end function dfun

    real(kind=real8) function  efun(x,y)
        real(kind=real8) :: x, y
        efun = 0.0_real8
        return
    end function efun

    real(kind=real8) function  ffun(x,y) 
        real(kind=real8) :: x, y
        ffun = 0.0_real8
        return
    end function ffun

    real(kind=real8) function BC(x,y) 
        real(kind=real8) :: x, y
        real(kind=real8), parameter :: half = 0.5_real8
        if (x == 0.0_real8) BC = y
        if (x == 1.0_real8) BC = half
        if (y == 0.0_real8) BC = sin((x-half)*(4.0_real8*pi))
        if (y == 1.0_real8) BC = cos((x-half)*(4.0_real8*pi))
        if ((x /= 0.0_real8) .and. (x /= 1.0_real8) .and.  &
            (y /= 0.0_real8) .and. (y /= 1.0_real8)) then
            write(stderr,666) 'Error: In function BC, got invalid arguments'
            write(stderr,777) 'x = ', x
            write(stderr,777) 'y = ', y
            stop
        end if
        return
        666 format(a)
        777 format(a, 2x, es27.17)
    end function BC
        
end module coeffs
