subroutine discretizer(meshpoints, nrows, ncols, nnz, rows, cols, vals, rhs, job)
!--------------------------------------------------------------------------------
! Set up the COO data structure for the five-point centered diff matrix, and the
! right hand side vector rhs().
! 
! This assumes that meshpoints is the number of nodes corresponding to unknowns in
! both the x and y directions. So it does not include the boundary nodes, for which
! the values are all given by Dirichlet conditions in function BC() in module
! coeffs.f90.
!-----------------
! Randall Bramley
! Department of Computer Science
! Indiana University, Bloomington
!--------------------------------------------------------------------------------

    use kindlytypes, only: int4 => integer4, int8 => integer8, real8
    use iso_fortran_env, only: stdout => output_unit, &
                               stderr => error_unit
    use coeffs
    use noisy
    use directions
    implicit none

    real(kind=real8), parameter :: ONE  = 1.0_real8
    real(kind=real8), parameter :: ZERO = 0.0_real8
    integer(kind=int4) :: nrows, ncols, nnz, job
    integer(kind=int4), allocatable :: rows(:), cols(:)
    real(kind=real8), allocatable :: vals(:), rhs(:)
    real(kind=real8) :: h

    integer(kind=int4) stat_allocate
    integer(kind=int4) :: node, j, nxoff, nyoff, meshpoints
    integer(kind=int4) :: ix, iy
    real(kind=real8)   :: x, y, stencil(6)

    nrows = meshpoints*meshpoints  ! Total number of unknowns
    ncols = nrows
    nnz   = 5*nrows - 4*meshpoints

    ! If only sizes are needed, bail out now
    if (job == 0) return

    node  = 1           ! counter for variable number (1:n)
    j = 1               ! counter for nonzeros in matrix (1:nnz)
    nxoff = 1           ! offset between consecutive x entries in the matrix
    nyoff = meshpoints  ! offset between consecutive y entries in the matrix

    h = 1.0_real8/(meshpoints+1.0_real8);

    if (benoisy > 0) then
        write(stdout,fmt=2000)' discretizer: meshpoints = ', meshpoints
        write(stdout,fmt=2000)' discretizer: nrows = ', nrows
        write(stdout,fmt=2000)' discretizer: ncols = ', ncols
        write(stdout,fmt=2000)' discretizer: nnz = ', nnz
        write(stdout,fmt=2000)' discretizer: nyoff = ', nyoff
    end if

    !============================================
    ! Compute matrix A and right hand side rhs:
    !============================================

    ! Loop over mesh points, x varying fastest:
    yaxis: do iy = 1, meshpoints
        y = real(iy,kind=real8)/(meshpoints + ONE)
        xaxis: do ix = 1, meshpoints
            x = real(ix,kind=real8)/(meshpoints + ONE)
            call getsten(x,y,meshpoints, stencil)
            rhs(node)  = stencil(otherside)

        ! Away from bottom boundary?
        if (iy > 1) then
            cols(j) = node - nyoff
            vals(j) = stencil(down)
            rows(j) = node
            j = j + 1
        else
            rhs(node) = rhs(node) - stencil(down)*BC(x,ZERO)
        end if

        ! Away from left hand boundary?
        if (ix > 1) then
            cols(j) = node - nxoff
            vals(j) = stencil(left)
            rows(j) = node
            j = j + 1
        else
            rhs(node) = rhs(node) - stencil(left)*BC(ZERO,y)
        end if

        ! Central node in stencil = diagonal element of A
        cols(j) = node
        vals(j) = stencil(1)
        rows(j) = node
        j = j + 1

        ! Away from right hand boundary?
        if (ix < meshpoints) then
            cols(j) = node + nxoff
            vals(j) = stencil(right)
            rows(j) = node
            j  = j + 1
        else
            rhs(node) = rhs(node) - stencil(right)*BC(ONE,y)
        end if

        ! Away from top boundary?
        if (iy < meshpoints) then
            cols(j) = node + nyoff
            vals(j) = stencil(up)
            rows(j) = node
            j = j + 1
        else
            rhs(node) = rhs(node) - stencil(up)*BC(x,ONE)
        end if

        ! Increment variable counter:
        node = node + 1
        end  do xaxis ! ix loop
    end  do yaxis ! iy loop

    ! Total number of unknowns:
    nrows = node - 1

    ! Total number of nonzeros in coefficient matrix:
    nnz = j - 1

    if (benoisy > 1) then
        write(stdout,fmt=2000)' discretizer has nnz = ', nnz
        write(stdout,fmt=2000)' discretizer has nrows = ', nrows
    end if
    if (benoisy > 2) then
        write(stdout,fmt=5000)' discretizer has rows = ', rows
        write(stdout,fmt=5000)' discretizer has cols = ', cols
        write(stdout,fmt=6000)' discretizer has vals = ', vals
    end if

    return  
    1000 format(a)
    2000 format(a, i0)
    3000 format(a, i0, a)
    4000 format(a, es27.17)
    5000 format(a, *(/,3x, i0))  ! infinite repeat descriptor
    6000 format(a, *(/,3x, es27.17))

end subroutine discretizer
